from __future__ import annotations

from typing import Literal, Optional, Union

import pandas as pd

from earningspy.common.constants import (
    FINVIZ_EARNINGS_DATE_KEY,
    DAYS_TO_EARNINGS_KEY_CAPITAL,
    TICKER_KEY_CAPITAL,
    ABS_RET_KEY,
    EXP_RET_KEY,
    RF_KEY,
    MARK_EXP_KEY,
    CAPM_KEY,
    VIX_KEY,
    EARNING_VIX_KEY,
    CAR_KEY,
    BHAR_KEY,
    ALLOWED_WINDOWS,
    AVAILABLE_CHECK_COLUMNS,
    DATADATE_KEY
)
from earningspy.calendars.utils import days_left
from earningspy.inspectors.mixins import CARMixin, TimeSeriesMixin


class PEADInspector(CARMixin, TimeSeriesMixin):
    """
    Calendar should be passed using this method
    pd.read_csv('<field-name>.csv', index_col=0, parse_dates=True)
    """

    def __init__(
        self,
        calendar: Optional[pd.DataFrame] = None,
        price_history: Optional[pd.DataFrame] = None,
    ):

        self.calendar: pd.DataFrame = self._load_calendar(calendar)
        self.backup = self.calendar.copy()
        self.price_history: Optional[pd.DataFrame] = self._load_price_history(price_history)

        self.remaining_data: pd.DataFrame = self.calendar[
            ~(self.calendar[DAYS_TO_EARNINGS_KEY_CAPITAL] < -3)
        ]
        self.merged_data: Optional[pd.DataFrame] = None
        self.affected_rows: Optional[pd.DataFrame] = None


    def _load_calendar(self, calendar: Optional[pd.DataFrame]) -> pd.DataFrame:

        calendar[DAYS_TO_EARNINGS_KEY_CAPITAL] = calendar.apply(lambda row: days_left(row), axis=1)
        calendar = calendar.sort_values(DAYS_TO_EARNINGS_KEY_CAPITAL, ascending=False)

        return calendar
    
    def _sort_calendar(self) -> None:
        self.calendar = self.calendar.sort_values(DAYS_TO_EARNINGS_KEY_CAPITAL, ascending=False)


    def inspect(
        self,
        days: int = 3,
        dry_run: bool = False,
        reuse_timeseries: bool = False,
        post_earnings: bool = False,
        async_: bool = False,
    ) -> Union["PEADInspector", pd.DataFrame]:

        if days not in ALLOWED_WINDOWS:
            raise Exception(f'Invalid day range. Select from {ALLOWED_WINDOWS}')

        self.affected_rows = self._get_affected_rows(days, post_earnings=post_earnings)
        if dry_run:
            self.affected_rows = self.affected_rows.reset_index()
            self.affected_rows = self.affected_rows.set_index([FINVIZ_EARNINGS_DATE_KEY])
            return self.affected_rows

        self._process_windows_columns(days=days, reuse_timeseries=reuse_timeseries, async_=async_)
        self._get_earnings_vix()
        self._find_and_remove_duplicates()
        self._sort_calendar()

        return self

    def refresh(
        self,
        days: int = 3,
        dry_run: bool = False,
        reuse_timeseries: bool = False,
        check_column: Optional[str] = None,
        deep: bool = False,
        async_: bool = False,
    ) -> Union["PEADInspector", pd.DataFrame]:

        if days not in ALLOWED_WINDOWS:
            raise Exception(f'Invalid day range. Select from {ALLOWED_WINDOWS}')

        if not check_column or check_column not in AVAILABLE_CHECK_COLUMNS:
            raise Exception(f"Provide a column to check for NaNs to do the refresh, must be from this list {AVAILABLE_CHECK_COLUMNS}")
        
        self.affected_rows = self._get_affected_rows(days, check_column=check_column, deep=deep)
        if dry_run:
            self.affected_rows = self.affected_rows.reset_index()
            self.affected_rows = self.affected_rows.set_index([FINVIZ_EARNINGS_DATE_KEY])
            return self.affected_rows
        
        self._process_windows_columns(days=days, reuse_timeseries=reuse_timeseries, async_=async_)
        self._get_earnings_vix()
        self._find_and_remove_duplicates()

        return self
        

    def _process_windows_columns(
        self,
        days: int = 3,
        reuse_timeseries: bool = False,
        async_: bool = False,
    ) -> None:

        if not reuse_timeseries:
            self.price_history = self.fetch_price_history(assets=set(self.affected_rows.index.get_level_values(1).to_list()), async_=async_)

        self.calendar = self.calendar.reset_index()
        self.calendar = self.calendar.set_index([FINVIZ_EARNINGS_DATE_KEY, TICKER_KEY_CAPITAL])

        self._get_windows_abnormal_return(days=days)
        self._get_windows_risk_free_rate(days=days)
        self._get_windows_expected_return(days=days)
        self._get_windows_market_expected_return(days=days)
        self._get_windows_capm(days=days)
        self._get_windows_car(days=days)
        self._get_windows_bhar(days=days)
        self._get_windows_vix(days=days)


    def _get_affected_rows(
        self,
        days: int,
        check_column: str = ABS_RET_KEY,
        deep: bool = False,
        post_earnings: bool = False,
    ) -> pd.DataFrame:

        start = days
        end = days + 30

        if deep:
            affected_rows = self.calendar[(self.calendar[DAYS_TO_EARNINGS_KEY_CAPITAL] <= -start)]
        else:
            affected_rows = self.calendar[(self.calendar[DAYS_TO_EARNINGS_KEY_CAPITAL] <= -start) &
                                          (self.calendar[DAYS_TO_EARNINGS_KEY_CAPITAL] >= -end)]

        if not post_earnings:
            try:
                affected_rows = affected_rows[affected_rows[check_column.format(days)].isna()]
            except KeyError:
                raise Exception(f"Check column {check_column.format(days)} not found in the calendar. Is this post earnings data?"
                       " If so, set post_earnings=True.")
        affected_rows = affected_rows.reset_index()
        affected_rows = affected_rows.set_index([FINVIZ_EARNINGS_DATE_KEY, TICKER_KEY_CAPITAL])
        return affected_rows


    def join(
        self,
        storage: pd.DataFrame,
        earnings_phase: Literal["pre", "post"] = "pre",
        preserve: Literal["canonical", "incoming"] = "canonical",
        future_threshold_days: int = 7,
        past_threshold_days: int = 1,
    ) -> pd.DataFrame:
        """
        Join calendar with storage data.

        :param storage: DataFrame with historical data
        :param earnings_phase: 'pre' or 'post'
        :param preserve: 'canonical' or 'incoming'
            canonical -> preserve data from the canonical calendar (self.calendar)
            incoming -> preserve data from storage
        """

        if storage is None or storage.empty:
            raise Exception("storage can't be empty")

        if self.calendar.empty:
            raise Exception("calendar is empty nothing to concat")

        if earnings_phase not in {"pre", "post"}:
            raise ValueError("earnings_phase must be either 'pre' or 'post'")

        if preserve not in {"canonical", "incoming"}:
            raise ValueError("preserve must be either 'canonical' or 'incoming'")

        if earnings_phase == "pre":
            # Do not get data that with more than 5 days before earnings
            storage = storage[(storage[DAYS_TO_EARNINGS_KEY_CAPITAL] > 0) & 
                              (storage[DAYS_TO_EARNINGS_KEY_CAPITAL] <= future_threshold_days)].copy()
            
        else:
            storage = storage[(storage[DAYS_TO_EARNINGS_KEY_CAPITAL] < -1) & 
                              (storage[DAYS_TO_EARNINGS_KEY_CAPITAL] >= -past_threshold_days)].copy()

        if storage.empty:
            print("No data to merge after filtering by earnings_phase")
            return self.calendar

        cal = self.calendar.copy()

        # --- Critical fix: ensure dedupe keys exist as COLUMNS (not only in the index) ---
        needed_keys = {FINVIZ_EARNINGS_DATE_KEY, TICKER_KEY_CAPITAL}

        if not needed_keys.issubset(storage.columns):
            storage = storage.reset_index()

        if not needed_keys.issubset(cal.columns):
            cal = cal.reset_index()

        # Ensure DATADATE exists in both frames and is consistently datetime
        if DATADATE_KEY not in storage.columns:
            storage[DATADATE_KEY] = pd.NaT
        if DATADATE_KEY not in cal.columns:
            cal[DATADATE_KEY] = pd.NaT

        storage[DATADATE_KEY] = pd.to_datetime(storage[DATADATE_KEY], errors="raise", format="mixed")
        cal[DATADATE_KEY] = pd.to_datetime(cal[DATADATE_KEY], errors="raise", format="mixed")

        # Preserve rule: canonical => calendar wins; incoming => storage wins
        if preserve == "canonical":
            cal["_src_pri"] = 1
            storage["_src_pri"] = 0
        else:  # preserve == "incoming"
            cal["_src_pri"] = 0
            storage["_src_pri"] = 1

        merged = pd.concat([storage, cal], join="outer", ignore_index=True)

        self.merged_data = (
            merged
            .sort_values(["_src_pri", DATADATE_KEY], ascending=[False, False], kind="mergesort")
            .drop_duplicates(subset=[FINVIZ_EARNINGS_DATE_KEY, TICKER_KEY_CAPITAL], keep="first")
            .drop(columns=["_src_pri"])
            .sort_values(DATADATE_KEY, ascending=False)
            .set_index(FINVIZ_EARNINGS_DATE_KEY)
        )

        self.merged_data[DAYS_TO_EARNINGS_KEY_CAPITAL] = (
            self.merged_data.apply(lambda row: days_left(row), axis=1)
        )
        self.merged_data = self.merged_data.sort_values(DAYS_TO_EARNINGS_KEY_CAPITAL, ascending=False)

        return self.merged_data


    def _get_windows_abnormal_return(self, days: int) -> None:

        label = ABS_RET_KEY.format(days)
    
        self.calendar.loc[self.affected_rows.index, label] = self.calendar.loc[self.affected_rows.index].apply(
            lambda row: self.get_window_pct_change(row, days=days), axis=1)

    
    def _get_windows_market_expected_return(self, days: int) -> None:
        
        label = MARK_EXP_KEY.format(days)
        self.calendar.loc[self.affected_rows.index, label] = self.calendar.loc[self.affected_rows.index].apply(
            lambda row: self.get_market_expected_return(row, days=days), axis=1)


    def _get_windows_capm(self, days: int) -> None:

        label = CAPM_KEY.format(days)
        self.calendar.loc[self.affected_rows.index, label] = self.calendar.loc[self.affected_rows.index].apply(
            lambda row: self.get_capm(row, days=days), axis=1)


    def _get_windows_expected_return(self, days: int) -> None:

        label = EXP_RET_KEY.format(days)
        self.calendar.loc[self.affected_rows.index, label] = self.calendar.loc[self.affected_rows.index].apply(
            lambda row: self.get_expected_return(row, days=days), axis=1)


    def _get_windows_risk_free_rate(self, days: int) -> None:
        label = RF_KEY.format(days)
        self.calendar.loc[self.affected_rows.index, label] = self.calendar.loc[self.affected_rows.index].apply(
            lambda row: self.get_risk_free_rate(row, days=days), axis=1)


    def _get_windows_vix(self, days: int) -> None:

        label = VIX_KEY.format(days)
        self.calendar.loc[self.affected_rows.index, label] = self.calendar.loc[self.affected_rows.index].apply(
            lambda row: self.get_vix(row, days=days), axis=1)


    def _get_earnings_vix(self) -> None:

        self.calendar.loc[self.affected_rows.index, EARNING_VIX_KEY] = self.calendar.loc[self.affected_rows.index].apply(
            lambda row: self.get_vix_for_date(row), axis=1)


    def _get_windows_car(self, days: int) -> None:
        label = CAR_KEY.format(days)
        ret_label = ABS_RET_KEY.format(days)
        capm_label = CAPM_KEY.format(days)
        self.calendar.loc[self.affected_rows.index, label] = (self.calendar[ret_label] - self.calendar[capm_label]).round(4)


    def _get_windows_bhar(self, days: int) -> None:
        label = BHAR_KEY.format(days)
        ret_label = ABS_RET_KEY.format(days)
        benchmark_label = MARK_EXP_KEY.format(days)
        self.calendar.loc[self.affected_rows.index, label] = (self.calendar[ret_label] - self.calendar[benchmark_label]).round(4)


    def _find_and_remove_duplicates(self) -> None:
        self.calendar = self.calendar.reset_index()
        self.calendar = self.calendar.set_index([FINVIZ_EARNINGS_DATE_KEY, TICKER_KEY_CAPITAL])
        self.calendar = self.calendar[~self.calendar.index.duplicated(keep='first')]

        self._remove_duplicate_ticker_quarters()

        self.calendar = self.calendar.reset_index()
        self.calendar = self.calendar.set_index(FINVIZ_EARNINGS_DATE_KEY)


    def _remove_duplicate_ticker_quarters(self) -> None:
        self.calendar = self.calendar.reset_index()
        self.calendar['year'] = self.calendar[FINVIZ_EARNINGS_DATE_KEY].dt.year
        self.calendar['quarter'] = self.calendar[FINVIZ_EARNINGS_DATE_KEY].dt.quarter
        self.calendar = self.calendar.sort_values(FINVIZ_EARNINGS_DATE_KEY)
        original_len = len(self.calendar)
        self.calendar = self.calendar.drop_duplicates(subset=[TICKER_KEY_CAPITAL, 'year', 'quarter'], keep='last')
        num_duplicates = original_len - len(self.calendar)
        print(f"Found and removed {num_duplicates} duplicate ticker-quarter entries.")
        self.calendar = self.calendar.drop(columns=['year', 'quarter'])
        self.calendar = self.calendar.set_index([FINVIZ_EARNINGS_DATE_KEY, TICKER_KEY_CAPITAL])



import asyncio
import aiohttp
import random

from datetime import datetime as dt
from dateutil.relativedelta import relativedelta
import pandas as pd
from tqdm import tqdm

# Existing functions (get_range, get_range_timestamps, prepare_data) remain unchanged
def get_range(range_from, end_date):

    accepted_values = ['3m', '9m', '1y', '5y', '10y']
    if range_from not in accepted_values:
        raise Exception('Invalid from value')

    if range_from == '3m':
        start_date = end_date - relativedelta(months=3)
    if range_from == '9m':
        start_date = end_date - relativedelta(months=9)
    if range_from == '1y':
        start_date = end_date - relativedelta(years=1)
    if range_from == '5y':
        start_date = end_date - relativedelta(years=5)
    if range_from == '10y':
        start_date = end_date - relativedelta(years=10)
    
    return str(start_date), str(end_date)


def get_range_timestamps(start_date, end_date):

    start_date = str(dt.strptime(start_date, "%Y-%m-%d")
        .timestamp()) \
        .replace('.0', '')

    end_date = str(dt.strptime(end_date, "%Y-%m-%d")
        .timestamp()) \
        .replace('.0', '')

    return start_date, end_date


def _normalize_close_frame(close_data, asset):
    if close_data.index.name != 'Date':
        close_data.index.name = 'Date'

    if asset not in close_data.columns:
        raise KeyError(asset)

    close_frame = close_data[[asset]].copy()
    close_frame.index = pd.to_datetime(close_frame.index, errors='coerce')
    close_frame = close_frame[~close_frame.index.isna()]
    close_frame = close_frame[~close_frame.index.duplicated(keep='last')]
    close_frame.index.name = 'Date'
    return close_frame


def _finalize_portfolio(close_frames):
    portfolio = pd.concat(close_frames, axis=1, join='outer')
    portfolio.index = pd.to_datetime(portfolio.index, errors='coerce')
    portfolio = portfolio[~portfolio.index.isna()]
    portfolio = portfolio[~portfolio.index.duplicated(keep='last')]
    portfolio.index.name = 'Date'
    portfolio = portfolio.sort_index()
    return portfolio.round(3)

async def get_one_ticker_async(session, asset, from_='3m', start_date=None, end_date=dt.now().date()):
    headers = {
        'User-Agent': "Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.2; .NET CLR 1.0.3705;)",
    }

    if start_date:
        end_date = str(end_date)
        start, end = get_range_timestamps(start_date, end_date)
    else:
        start, end = get_range(range_from=from_, end_date=end_date)
        start, end = get_range_timestamps(start, end)

    url = f"https://query2.finance.yahoo.com/v8/finance/chart/{asset}?period1={start}&period2={end}&interval=1d&events=history"
    
    try:
        async with session.get(url, headers=headers) as response:
            if response.status == 200:
                data = await response.json()
                df = pd.DataFrame.from_dict(data['chart']['result'][0]['indicators']['quote'][0])
                df['Date'] = data['chart']['result'][0]['timestamp']
                df['Date'] = df['Date'].apply(dt.fromtimestamp)
                df = df.set_index('Date', drop=True)
                df.index = df.index.normalize()
                return asset, df.round(2)
            else:
                # print(f"Could not retrieve data for {asset}: {response.status}")
                return asset, None
    except Exception as e:
        print(f"Error fetching data for {asset}: {str(e)}")
        return asset, None

async def get_portfolio_async(assets, from_='3m', start_date=None, end_date=dt.now().date()):
    close_frames = {}
    not_found = []
    seen_assets = set()
    assets = list(assets)

    async with aiohttp.ClientSession() as session:
        tasks = []
        for asset in assets:
            tasks.append(asyncio.ensure_future(get_one_ticker_async(session, asset, from_=from_, start_date=start_date, end_date=end_date)))
        
        for future in tqdm(asyncio.as_completed(tasks), total=len(assets)):
            asset, ticker_data = await future
            if ticker_data is None:
                not_found.append(asset)
                continue
            
            try:
                close_data = prepare_data(ticker_data, asset)
                close_frame = _normalize_close_frame(close_data, asset)
            except (KeyError, TypeError, ValueError):
                not_found.append(asset)
                continue

            if close_frame.empty or asset in seen_assets:
                continue

            close_frames[asset] = close_frame
            seen_assets.add(asset)
            
            # Random sleep to avoid rate limiting
            await asyncio.sleep(random.uniform(0.5, 1.0))

    if not close_frames:
        raise ValueError("No valid assets found — portfolio is empty")

    portfolio = _finalize_portfolio([close_frames[asset] for asset in assets if asset in close_frames])
    if len(not_found):
        print(f"Not found assets: {len(set(not_found))}, {set(not_found)}")
    return portfolio

def prepare_data(data, ticker):
    data = data.drop(['open', 'high', 'low', 'volume'], axis=1)
    data = data.rename(columns={"close": ticker})
    return data.round(2)


def get_portfolio(assets, from_='3m', start_date=None, end_date=dt.now().date()):
    """
    Synchronous wrapper for the asynchronous get_portfolio_async function.
    """
    loop = asyncio.get_event_loop()
    portfolio = loop.run_until_complete(get_portfolio_async(assets, from_=from_, start_date=start_date, end_date=end_date))
    return portfolio



import csv
import io
import re
import sqlite3


def create_connection(sqlite_file):
    """ Creates a database connection. """

    try:
        conn = sqlite3.connect(sqlite_file)
        return conn
    except sqlite3.Error as error:
        raise (
            "An error has occurred while connecting to the database: ",
            error.args[0],
        )


def __write_csv_to_stream(stream, headers, data):
    """Writes the data in CSV format to a stream."""

    dict_writer = csv.DictWriter(stream, headers)
    dict_writer.writeheader()
    dict_writer.writerows(data)


def export_to_csv(headers, data, filename=None, mode="w", newline=""):
    """Exports the generated table into a CSV file if a file is mentioned.
    Returns the CSV table as a string if no file is mentioned."""

    if filename:
        with open(filename, mode, newline=newline) as output_file:
            __write_csv_to_stream(output_file, headers, data)
        return None
    stream = io.StringIO()
    __write_csv_to_stream(stream, headers, data)
    return stream.getvalue()


def export_to_db(headers, data, filename):
    """ Exports the generated table into a SQLite database into a file."""

    field_list = ""
    table_name = "screener_results"  # name of the table to be created
    conn = create_connection(filename)
    c = conn.cursor()

    for field in headers:
        field_cleaned = re.sub(r"[^\w\s]", "", field)
        field_cleaned = field_cleaned.replace(" ", "")
        field_cleaned = field_cleaned.replace("50DHigh", "High50D")
        field_cleaned = field_cleaned.replace("50DLow", "Low50D")
        field_cleaned = field_cleaned.replace("52WHigh", "High52W")
        field_cleaned = field_cleaned.replace("52WLow", "Low52W")
        field_list += field_cleaned + " TEXT, "

    c.execute(f"CREATE TABLE IF NOT EXISTS {table_name} ({field_list[:-2]});")

    inserts = ""
    for row in data:

        insert_fields = "("
        for field, value in row.items():

            insert_fields += '"' + value + '", '

        inserts += insert_fields[:-2] + "), "

    insert_lines = inserts[:-2]

    try:
        c.execute(f"INSERT INTO {table_name} VALUES {insert_lines};")
    except sqlite3.Error as error:
        print("An error has occurred", error.args[0])

    conn.commit()
    conn.close()



from __future__ import annotations

import math
from typing import Iterable, Optional, Union

import numpy as np
import pandas as pd
from pandas.tseries.offsets import BDay

from earningspy.generators.yahoo.async_timeseries import get_portfolio as async_get_portfolio
from earningspy.generators.yahoo.time_series import get_portfolio
from earningspy.common.constants import (
    MARKET_DATA_TICKERS,
    TBILL_10_YEAR,
    EXP_RET_KEY,
    SP_500_TICKER,
    BETA_KEY,
    RF_KEY,
    TBILL_10_YEAR,
    VIX_TICKER,
)

class CARMixin:

    def _coerce_price_timestamp(self, value: Union[pd.Timestamp, str]) -> pd.Timestamp:
        return pd.Timestamp(value).normalize()

    def _resolve_price_date(
        self,
        target: Union[pd.Timestamp, str],
        method: str = "nearest",
        fallback_method: Optional[str] = None,
    ) -> pd.Timestamp:
        target_ts = self._coerce_price_timestamp(target)

        if target_ts in self.price_history.index:
            return target_ts

        locator = self.price_history.index.get_indexer([target_ts], method=method)[0]
        if locator == -1 and fallback_method is not None:
            locator = self.price_history.index.get_indexer([target_ts], method=fallback_method)[0]
        if locator == -1:
            raise KeyError(f"Could not resolve trading date for {target_ts}")

        return self.price_history.index[locator]

    def _resolve_window_dates(self, earnings_date: pd.Timestamp, days: int) -> tuple[pd.Timestamp, pd.Timestamp]:
        earnings_ts = self._coerce_price_timestamp(earnings_date)
        initial_target = earnings_ts - BDay(1)
        end_target = earnings_ts + BDay(days)

        # Start should not jump forward when a prior trading day exists.
        initial_date = self._resolve_price_date(initial_target, method="pad", fallback_method="backfill")
        # End should not jump backward when a later trading day exists.
        end_date = self._resolve_price_date(end_target, method="backfill", fallback_method="pad")

        return initial_date, end_date

    def get_window_pct_change(self, row: pd.Series, days: int) -> float:
        earnings_date = row.name[0]
        ticker = row.name[1]

        initial_target = self._coerce_price_timestamp(earnings_date) - BDay(1)
        end_target = self._coerce_price_timestamp(earnings_date) + BDay(days)

        try:
            initial_date, end_date = self._resolve_window_dates(earnings_date, days)
            ts_slice = self.price_history.loc[:end_date, ticker].copy()
            ts_slice_ffill = ts_slice.ffill().loc[initial_date:end_date]
            ts_valid = ts_slice_ffill.dropna()

            if len(ts_valid) < 2:
                print(
                    "⚠️ Not enough data after ffill for "
                    f"{ticker}. target window {initial_target} -> {end_target}, "
                    f"resolved window {initial_date} -> {end_date}"
                )
                return np.nan

            start_price = ts_valid.iloc[0]
            end_price = ts_valid.iloc[-1]

            pct = (end_price - start_price) / start_price
            return np.round(pct, 4)

        except KeyError as exc:
            print(f"❌ Could not compute window pct change for {ticker}: {exc}")
            return np.nan


    def get_capm(self, row: pd.Series, days: int = 0) -> float:

        rf_label = RF_KEY.format(days).strip()
        R_label = EXP_RET_KEY.format(days).strip()

        rf = row[rf_label]
        R = row[R_label]
        b = row[BETA_KEY]

        capm = rf + b * (R - rf)
        return np.round(capm, 4)


    def get_expected_return(self, row: pd.Series, days: int) -> float:

        date = row.name[0]
        ticker = row.name[1]

        try:
            if days == 3:
                exp_ret = self.price_history[ticker].loc[:date].pct_change(days, fill_method=None).mean()
            elif days == 30:
                exp_ret = self.price_history[ticker].loc[:date].resample('1ME').ffill().pct_change(fill_method=None).mean()
            elif days == 60:
                exp_ret = self.price_history[ticker].loc[:date].resample('2ME').ffill().pct_change(fill_method=None).mean()
        except KeyError:
            exp_ret = np.nan

        return np.round(exp_ret, 4)


    def get_market_expected_return(self, row: pd.Series, days: int) -> float:

        date = row.name[0]
        try:
            if days == 3:
                exp_ret = self.price_history[SP_500_TICKER].loc[:date].pct_change(days, fill_method=None).mean()
            elif days == 30:
                exp_ret = self.price_history[SP_500_TICKER].loc[:date].resample('1ME').ffill().pct_change(fill_method=None).mean()
            elif days == 60:
                exp_ret = self.price_history[SP_500_TICKER].loc[:date].resample('2ME').ffill().pct_change(fill_method=None).mean()
        except KeyError:
            exp_ret = np.nan

        return np.round(exp_ret, 4)


    def get_risk_free_rate(self, row: pd.Series, days: int) -> float:

        date = self._resolve_price_date(row.name[0], method="pad", fallback_method="backfill")
        rf = self.price_history.loc[date][TBILL_10_YEAR]

        if math.isnan(rf):
            rf = self.price_history[TBILL_10_YEAR].mean()
        rf = (rf / 100) * (days / 251)
        return np.round(rf, 4)


    def get_vix(self, row: pd.Series, days: int = 0) -> float:
        earnings_date = row.name[0]
        initial_date, end_date = self._resolve_window_dates(earnings_date, days)

        ts_slice = self.price_history.loc[initial_date:end_date]
        try:
            value = ts_slice[VIX_TICKER].mean()
        except KeyError as e:
            print(f"VIX windows data is not in timeseries data")
            value = np.nan

        return np.round(value, 2)


    def get_vix_for_date(self, row: pd.Series) -> float:
        earnings_date = row.name[0]
        ticker = row.name[1]

        if not pd.isna(row['IS_AMC']) and row['IS_AMC'] == 1:
            earnings_date = self._coerce_price_timestamp(earnings_date) - BDay(1)

        earnings_date = self._resolve_price_date(earnings_date, method="pad", fallback_method="backfill")

        try:
            value = self.price_history.loc[earnings_date][VIX_TICKER]
        except KeyError as e:
            print(f"VIX value not present for {ticker} is not in timeseries data")
            value = np.nan

        return np.round(value, 2)


class TimeSeriesMixin:

    def _load_price_history(
        self,
        price_history: Optional[pd.DataFrame],
        assets: Optional[Iterable[str]] = None,
    ) -> Optional[pd.DataFrame]:

        if price_history is None or price_history.empty:
            return None
        
        print("timeseries found")
        price_history.index = pd.to_datetime(price_history.index, errors='coerce')
        price_history = price_history[~price_history.index.isna()]
        price_history = price_history[~price_history.index.duplicated(keep='first')]        
        price_history = price_history.sort_index()
    
        return price_history

    def fetch_price_history(
        self,
        assets: Iterable[str],
        from_: str = "5y",
        async_: bool = False,
    ) -> pd.DataFrame:

        market_assets = MARKET_DATA_TICKERS

        if async_:
            price_history = async_get_portfolio(list(assets) + market_assets, from_=from_)
        else:
            price_history = get_portfolio(list(assets) + market_assets, from_=from_)        
        price_history.index = pd.to_datetime(price_history.index, errors='coerce')
        price_history = price_history[~price_history.index.isna()]
        price_history = price_history[~price_history.index.duplicated(keep='first')]        
        price_history = price_history.sort_index()

        return price_history



import json
import pathlib
import urllib.request
from urllib.parse import parse_qs as urlparse_qs
from urllib.parse import urlencode, urlparse

from bs4 import BeautifulSoup
from user_agent import generate_user_agent

import earningspy.generators.finviz.helper_functions.scraper_functions as scrape
from earningspy.generators.finviz.helper_functions.display_functions import create_table_string
from earningspy.generators.finviz.helper_functions.error_handling import InvalidTableType, NoResults
from earningspy.generators.finviz.helper_functions.request_functions import (Connector,
                                                       http_request_get,
                                                       sequential_data_scrape)
from earningspy.generators.finviz.helper_functions.save_data import export_to_csv, export_to_db

TABLE_TYPES = {
    "Overview": "111",
    "Valuation": "121",
    "Ownership": "131",
    "Performance": "141",
    "Custom": "152",
    "Financial": "161",
    "Technical": "171",
}


class Screener(object):
    """ Used to download data from https://www.finviz.com/screener.ashx. """

    @classmethod
    def init_from_url(cls, url, rows=None):
        """
        Initializes from url

        :param url: screener url
        :type url: string
        :param rows: total number of rows to get
        :type rows: int
        """

        split_query = urlparse_qs(urlparse(url).query)

        tickers = split_query["t"][0].split(",") if "t" in split_query else None
        filters = split_query["f"][0].split(",") if "f" in split_query else None
        custom = split_query["c"][0].split(",") if "c" in split_query else None
        order = split_query["o"][0] if "o" in split_query else ""
        signal = split_query["s"][0] if "s" in split_query else ""

        table = "Overview"
        if "v" in split_query:
            table_numbers_types = {v: k for k, v in TABLE_TYPES.items()}
            table_number_string = split_query["v"][0][0:3]
            try:
                table = table_numbers_types[table_number_string]
            except KeyError:
                raise InvalidTableType(split_query["v"][0])

        return cls(tickers, filters, rows, order, signal, table, custom)

    def __init__(
        self,
        tickers=None,
        filters=None,
        rows=None,
        order="",
        signal="",
        table=None,
        custom=None,
        user_agent=generate_user_agent(),
        request_method="sequential",
    ):
        """
        Initializes all variables to its values

        :param tickers: collection of ticker strings eg.: ['AAPL', 'AMD', 'WMT']
        :type tickers: list
        :param filters: collection of filters strings eg.: ['exch_nasd', 'idx_sp500', 'fa_div_none']
        :type filters: list
        :param rows: total number of rows to get
        :type rows: int
        :param order: table order eg.: '-price' (to sort table by descending price)
        :type order: str
        :param signal: show by signal eg.: 'n_majornews' (for stocks with major news)
        :type signal: str
        :param table: table type eg.: 'Performance'
        :type table: str
        :param custom: collection of custom columns eg.: ['1', '21', '23', '45']
        :type custom: list
        :var self.data: list of dictionaries containing row data
        :type self.data: list
        """

        if tickers is None:
            self._tickers = []
        else:
            self._tickers = tickers

        if filters is None:
            self._filters = []
        else:
            self._filters = filters

        if table is None:
            self._table = "111"
        else:
            self._table = self.__check_table(table)

        if custom is None:
            self._custom = []
        else:
            self._table = "152"
            self._custom = custom

            if (
                "0" not in self._custom
            ):  # 0 (No.) is required for the sequence algorithm to work
                self._custom = ["0"] + self._custom

        self._rows = rows
        self._order = order
        self._signal = signal
        self._user_agent = user_agent
        self._request_method = request_method

        self.analysis = []
        self.data = self.__search_screener()

    def __call__(
        self,
        tickers=None,
        filters=None,
        rows=None,
        order="",
        signal="",
        table=None,
        custom=None,
    ):
        """
        Adds more filters to the screener. Example usage:

        stock_list = Screener(filters=['cap_large'])  # All the stocks with large market cap
        # After analyzing you decide you want to see which of the stocks have high dividend yield
        # and show their performance:
        stock_list(filters=['fa_div_high'], table='Performance')
        # Shows performance of stocks with large market cap and high dividend yield
        """

        if tickers:
            [self._tickers.append(item) for item in tickers]

        if filters:
            [self._filters.append(item) for item in filters]

        if table:
            self._table = self.__check_table(table)

        if order:
            self._order = order

        if signal:
            self._signal = signal

        if rows:
            self._rows = rows

        if custom:
            self._custom = custom

        self.analysis = []
        self.data = self.__search_screener()

    add = __call__

    def __str__(self):
        """ Returns a readable representation of a table. """

        table_list = [self.headers]

        for row in self.data:
            table_list.append([row[col] or "" for col in self.headers])

        return create_table_string(table_list)

    def __repr__(self):
        """ Returns a string representation of the parameter's values. """

        values = (
            f"tickers: {tuple(self._tickers)}\n"
            f"filters: {tuple(self._filters)}\n"
            f"rows: {self._rows}\n"
            f"order: {self._order}\n"
            f"signal: {self._signal}\n"
            f"table: {self._table}\n"
            f"table: {self._custom}"
        )

        return values

    def __len__(self):
        """ Returns an int with the number of total rows. """

        return int(self._rows)

    def __getitem__(self, position):
        """ Returns a dictionary containing specific row data. """

        return self.data[position]

    get = __getitem__

    @staticmethod
    def __check_table(input_table):
        """ Checks if the user input for table type is correct. Otherwise, raises an InvalidTableType error. """

        try:
            table = TABLE_TYPES[input_table]
            return table
        except KeyError:
            raise InvalidTableType(input_table)

    @staticmethod
    def load_filter_dict(reload=True):
        """
        Get dict of available filters. File containing json specification of filters will be built if it doesn't exist
        or if reload is False
        """

        # Get location of filter.json
        json_directory = pathlib.Path(__file__).parent
        json_file = pathlib.Path.joinpath(json_directory, "filters.json")

        # Reload the filters JSON file if present and requested
        if reload and json_file.is_file():
            with open(json_file, "r") as fp:
                return json.load(fp)

        # Get html from main filter page, ft=4 ensures all filters are present
        hdr = {
            "User-Agent": "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.11 (KHTML, like Gecko) "
            "Chrome/23.0.1271.64 Safari/537.11"
        }
        url = "https://finviz.com/screener.ashx?ft=4"
        req = urllib.request.Request(url, headers=hdr)
        with urllib.request.urlopen(req) as response:
            html = response.read().decode("utf-8")

        # Parse html and locate table we are interested in.
        # Use one of the text values and get the parent table from that
        bs = BeautifulSoup(html, "html.parser")
        filters_table = None
        for td in bs.find_all("td"):
            if td.get_text().strip() == "Exchange":
                filters_table = td.find_parent("table")
        if filters_table is None:
            raise Exception("Could not locate filter parameters")

        # Delete all div tags, we don't need them
        for div in filters_table.find_all("div"):
            div.decompose()

        # Populate dict with filtering options and corresponding filter tags
        filter_dict = {}
        td_list = filters_table.find_all("td")

        for i in range(0, len(td_list) - 2, 2):
            current_dict = {}
            if td_list[i].get_text().strip() == "":
                continue

            # Even td elements contain filter name (as shown on web page)
            filter_text = td_list[i].get_text().strip()

            # Odd td elements contain the filter tag and options
            selections = td_list[i + 1].find("select")
            filter_name = selections.get("data-filter").strip()

            # Store filter options for current filter
            options = selections.find_all("option", {"value": True})
            for opt in options:
                # Encoded filter string
                value = opt.get("value").strip()

                # String shown in pull-down menu
                text = opt.get_text()

                # Filter out unwanted items
                if value is None or "Elite" in text:
                    continue

                # Make filter string and store in dict
                current_dict[text] = f"{filter_name}_{value}"

            # Store current filter dict
            filter_dict[filter_text] = current_dict

        # Save filter dict to finviz directory
        try:
            with open(json_file, "w") as fp:
                json.dump(filter_dict, fp)
        except Exception as e:
            print(e)
            print("Unable to write to file{}".format(json_file))

        return filter_dict

    def to_sqlite(self, filename):
        """Exports the generated table into a SQLite database.

        :param filename: SQLite database file path
        :type filename: str
        """

        export_to_db(self.headers, self.data, filename)

    def to_csv(self, filename: str):
        """Exports the generated table into a CSV file.
        Returns a CSV string if filename is None.

        :param filename: CSV file path
        :type filename: str
        """

        if filename and filename.endswith(".csv"):
            filename = filename[:-4]

        if len(self.analysis) > 0:
            export_to_csv(
                [
                    "ticker",
                    "date",
                    "category",
                    "analyst",
                    "rating",
                    "price_from",
                    "price_to",
                ],
                self.analysis,
                f"{filename}-analysts.csv",
            )

        return export_to_csv(self.headers, self.data, f"{filename}.csv")

    def get_charts(self, period="d", size="l", chart_type="c", ta="1"):
        """
        Downloads the charts of all tickers shown by the table.

        :param period: table period eg. : 'd', 'w' or 'm' for daily, weekly and monthly periods
        :type period: str
        :param size: table size eg.: 'l' for large or 's' for small - choose large for better quality but higher size
        :type size: str
        :param chart_type: chart type: 'c' for candles or 'l' for lines
        :type chart_type: str
        :param ta: technical analysis eg.: '1' to show ta '0' to hide ta
        :type ta: str
        """

        encoded_payload = urlencode(
            {"ty": chart_type, "ta": ta, "p": period, "s": size}
        )

        sequential_data_scrape(
            scrape.download_chart_image,
            [
                f"https://finviz.com/chart.ashx?{encoded_payload}&t={row.get('Ticker')}"
                for row in self.data
            ],
            self._user_agent,
        )

    def get_ticker_details(self):
        """
        Downloads the details of all tickers shown by the table.
        """

        ticker_data = sequential_data_scrape(
            scrape.download_ticker_details,
            [
                f"https://finviz.com/quote.ashx?&t={row.get('Ticker')}"
                for row in self.data
            ],
            self._user_agent,
        )

        for entry in ticker_data:
            for key, value in entry.items():
                for ticker_generic in self.data:
                    if ticker_generic.get("Ticker") == key:
                        if "Sales" not in self.headers:
                            self.headers.extend(list(value[0].keys()))

                        ticker_generic.update(value[0])
                        self.analysis.extend(value[1])

        return self.data
    
    def get_ticker_details_by_ticker(self, ticker):

        """
        Downloads the details of all tickers shown by the table.
        """

        ticker_data = sequential_data_scrape(
            scrape.download_ticker_details,
            [
                f"https://finviz.com/quote.ashx?&t={ticker}"
                for row in self.data
            ],
            self._user_agent,
        )

        return ticker_data

    def __check_rows(self):
        """
        Checks if the user input for row number is correct.
        Otherwise, modifies the number or raises NoResults error.
        """

        self._total_rows = scrape.get_total_rows(self._page_content)

        if self._total_rows == 0:
            raise NoResults(self._url.split("?")[1])
        elif self._rows is None or self._rows > self._total_rows:
            return self._total_rows
        else:
            return self._rows

    def __get_table_headers(self):
        """ Private function used to return table headers. """
        headers = []

        header_elements = self._page_content.cssselect('tr[valign="middle"]')[0].xpath("th")
        
        for header_element in header_elements:
            # Use normalize-space to extract text content while ignoring internal elements
            header_text = header_element.xpath("normalize-space()")
            
            if header_text:
                headers.append(header_text)
        
        return headers

    def __search_screener(self):
        """ Private function used to return data from the FinViz screener. """

        self._page_content, self._url = http_request_get(
            "https://finviz.com/screener.ashx",
            payload={
                "v": self._table,
                "t": ",".join(self._tickers),
                "f": ",".join(self._filters),
                "o": self._order,
                "s": self._signal,
                "c": ",".join(self._custom),
            },
            user_agent=self._user_agent,
        )

        self._rows = self.__check_rows()
        self.headers = self.__get_table_headers()

        if self._request_method == "async":
            async_connector = Connector(
                scrape.get_table,
                scrape.get_page_urls(self._page_content, self._rows, self._url),
                self._user_agent,
                self.headers,
                self._rows,
                css_select=True,
            )
            pages_data = async_connector.run_connector()
        else:
            pages_data = sequential_data_scrape(
                scrape.get_table,
                scrape.get_page_urls(self._page_content, self._rows, self._url),
                self._user_agent,
                self.headers,
                self._rows,
            )

        data = []
        for page in pages_data:
            for row in page:
                data.append(row)

        return data



import logging
from time import sleep
from datetime import datetime as dt

import pandas as pd
import requests
from dateutil.relativedelta import relativedelta
from tqdm import tqdm

# Fixing this problem: https://www.reddit.com/r/sheets/comments/1farvxr/broken_yahoo_finance_url/

logger = logging.getLogger(__name__)


def _get_response_hint(response, limit=200):
    response_text = str(getattr(response, 'text', '') or '').strip().replace('\n', ' ')
    if not response_text:
        return ''
    if len(response_text) > limit:
        return f"{response_text[:limit]}..."
    return response_text


def get_range(range_from, end_date):

    accepted_values = ['3m', '9m', '1y', '5y', '10y', '20y', '30y']
    if range_from not in accepted_values:
        raise Exception('Invalid from value')

    if range_from == '3m':
        start_date = end_date - relativedelta(months=3)
    if range_from == '9m':
        start_date = end_date - relativedelta(months=9)
    if range_from == '1y':
        start_date = end_date - relativedelta(years=1)
    if range_from == '5y':
        start_date = end_date - relativedelta(years=5)
    if range_from == '10y':
        start_date = end_date - relativedelta(years=10)
    if range_from == '20y':
        start_date = end_date - relativedelta(years=20)
    if range_from == '30y':
        start_date = end_date - relativedelta(years=30)

    return str(start_date), str(end_date)


def get_range_timestamps(start_date, end_date):

    start_date = str(dt.strptime(start_date, "%Y-%m-%d")
        .timestamp()) \
        .replace('.0', '')

    end_date = str(dt.strptime(end_date, "%Y-%m-%d")
        .timestamp()) \
        .replace('.0', '')

    return start_date, end_date


def _normalize_close_frame(close_data, asset):
    if close_data.index.name != 'Date':
        close_data.index.name = 'Date'

    if asset not in close_data.columns:
        raise KeyError(asset)

    close_frame = close_data[[asset]].copy()
    close_frame.index = pd.to_datetime(close_frame.index, errors='coerce')
    close_frame = close_frame[~close_frame.index.isna()]
    close_frame = close_frame[~close_frame.index.duplicated(keep='last')]
    close_frame.index.name = 'Date'
    return close_frame


def _finalize_portfolio(close_frames):
    portfolio = pd.concat(close_frames, axis=1, join='outer')
    portfolio.index = pd.to_datetime(portfolio.index, errors='coerce')
    portfolio = portfolio[~portfolio.index.isna()]
    portfolio = portfolio[~portfolio.index.duplicated(keep='last')]
    portfolio.index.name = 'Date'
    portfolio = portfolio.sort_index()
    return portfolio.round(3)


def get_portfolio(assets, from_='3m', start_date=None, end_date=dt.now().date()):
    close_frames = []
    not_found = []
    seen_assets = set()
    assets = list(assets)

    logger.info(
        "Fetching Yahoo portfolio for %s assets with from_=%s start_date=%s end_date=%s",
        len(assets),
        from_,
        start_date,
        end_date,
    )

    for asset in tqdm(assets):
        ticker_data = get_one_ticker(asset, from_=from_, start_date=start_date, end_date=end_date)
        if ticker_data is None:
            logger.warning("Skipping %s: no data returned from Yahoo fetch", asset)
            not_found.append(asset)
            continue

        if ticker_data.empty:
            logger.warning("Skipping %s: Yahoo returned an empty dataset", asset)
            not_found.append(asset)
            continue

        try:
            close_data = prepare_data(ticker_data, asset)
        except (KeyError, TypeError, ValueError) as exc:
            logger.warning("Skipping %s: could not prepare ticker data (%s)", asset, exc)
            not_found.append(asset)
            continue

        try:
            close_frame = _normalize_close_frame(close_data, asset)
        except (KeyError, TypeError, ValueError) as exc:
            logger.warning("Skipping %s: malformed data after preparation (%s)", asset, exc)
            not_found.append(asset)
            continue
        except Exception:
            logger.exception("Skipping %s: could not normalize prepared data", asset)
            not_found.append(asset)
            continue

        if close_frame.empty:
            logger.warning("Skipping %s: malformed data after preparation", asset)
            not_found.append(asset)
            continue

        if asset in seen_assets:
            continue

        close_frames.append(close_frame)
        seen_assets.add(asset)

        sleep(1)

    if not close_frames:
        logger.error(
            "No valid assets found for Yahoo portfolio request. requested_assets=%s failure_count=%s",
            assets,
            len(not_found),
        )
        raise ValueError("No valid assets found — portfolio is empty")


    portfolio = _finalize_portfolio(close_frames)

    logger.info("Not found assets: %s, %s", len(set(not_found)), sorted(set(not_found)))
    return portfolio


def get_one_ticker(asset, from_='3m', start_date=None, end_date=dt.now().date()):
    
    headers = {
        'User-Agent': "Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.2; .NET CLR 1.0.3705;)",
    }

    if start_date:
        end_date = str(end_date)
        start, end = get_range_timestamps(start_date, end_date)
    else:
        start, end = get_range(range_from=from_, end_date=end_date)
        start, end = get_range_timestamps(start, end)

    url = f"https://query2.finance.yahoo.com/v8/finance/chart/{asset}?" \
          f"period1={start}&period2={end}&interval=1d&events=history"

    logger.debug(
        "Fetching Yahoo ticker asset=%s from_=%s start_date=%s end_date=%s start_ts=%s end_ts=%s url=%s",
        asset,
        from_,
        start_date,
        end_date,
        start,
        end,
        url,
    )

    try:
        response = requests.get(url, headers=headers)
    except requests.RequestException as exc:
        logger.warning("Yahoo request failed for %s: %s", asset, exc)
        return None
    except Exception:
        logger.exception("Unexpected Yahoo request failure for %s", asset)
        return None

    if not response.ok:
        logger.warning(
            "Yahoo request returned non-OK response for %s: status=%s body_hint=%s",
            asset,
            getattr(response, 'status_code', 'unknown'),
            _get_response_hint(response),
        )
        return None

    try:
        payload = response.json()
        result = payload['chart']['result'][0]
        data = pd.DataFrame.from_dict(result['indicators']['quote'][0])
        data['Date'] = result['timestamp']
        data['Date'] = data['Date'].apply(dt.fromtimestamp)
    except (ValueError, KeyError, IndexError, TypeError) as exc:
        logger.warning("Yahoo payload parsing failed for %s: %s", asset, exc)
        return None
    except Exception:
        logger.exception("Unexpected Yahoo payload parsing failure for %s", asset)
        return None

    data = data.set_index('Date', drop=True)
    data.index = data.index.normalize()

    data.index.name = 'Date'

    return data.round(2)

    
def prepare_data(data, ticker):
    data = data.drop(['open', 'high', 'low', 'volume'], axis=1)
    data = data.rename(columns={"close": ticker})
    return data.round(2)



import os
import json
import pytest


def screener_mock_data():
    """Load screener_mock.json and return a list of dicts."""
    here = os.path.dirname(__file__)
    json_path = os.path.join(here, './screener_mock.json')
    with open(json_path, 'r') as f:
        data = json.load(f)
    return data



import pandas as pd
from earningspy.generators.finviz.screener import Screener
from pprint import pprint as pp
from earningspy.generators.finviz.constants import (
    CUSTOM_TABLE_ALL_FIELDS_NEW,    
    CUSTOM_TABLE_FIELDS_ON_URL,
    TICKER_KEY,
    VALID_SCOPES_EARNING_SCOPES
)
from earningspy.generators.finviz.utils import finviz_data_preprocessor


FINVIZ_URL = "https://finviz.com/screener.ashx?v=152&f={}{}&o={}"


def get_filters(sub_category=None, raw=False):
    filters = Screener.load_filter_dict()
    if raw:
        pp(Screener.load_filter_dict())
        return
    if not sub_category:
        for category in Screener.load_filter_dict().keys():
            print(f'{category}')
        return
    return filters.get(sub_category)


def _get_screener_data(filters=None, order='marketcap', query=None, print_url=False):

    if not query:
        query = FINVIZ_URL.format(filters, CUSTOM_TABLE_FIELDS_ON_URL, order)
    if print_url:
        print(query)
    stock_list = Screener.init_from_url(query)
    data = pd.DataFrame(index=CUSTOM_TABLE_ALL_FIELDS_NEW)
    for stock in stock_list:
        ticker = stock.get(TICKER_KEY)
        ticker_data = pd.DataFrame(index=CUSTOM_TABLE_ALL_FIELDS_NEW)
        for key, value in stock.items():
            if key in CUSTOM_TABLE_ALL_FIELDS_NEW:
                ticker_data.loc[key, ticker] = value
        data = pd.concat([data, ticker_data], axis=1)
    return finviz_data_preprocessor(data)


def get_by_earnings_date(scope, print_url=False):
    if scope not in VALID_SCOPES_EARNING_SCOPES:
        raise Exception(f"Invalid scope. Use {VALID_SCOPES_EARNING_SCOPES} instead")
    
    if scope == 'last_week':
        filters = 'earningsdate_prevweek'
    elif scope == 'this_week':
        filters = 'earningsdate_thisweek'
    elif scope == 'next_week':
        filters = 'earningsdate_nextweek'
    elif scope == 'today_bmo':
        filters = 'earningsdate_todaybefore'
    elif scope == 'yesterday_amc':
        filters = 'earningsdate_yesterdayafter'
    elif scope == 'today':
        filters = 'earningsdate_today'
    elif scope == 'this_month':
        filters = 'earningsdate_thismonth'

    return _get_screener_data(filters, print_url=print_url)


def get_by_tickers(tickers, order='marketcap'):
    tickers = ','.join(tickers)
    ticker_query = f'https://finviz.com/screener.ashx?t={tickers}' + CUSTOM_TABLE_FIELDS_ON_URL + f"&o={order}"
    return _get_screener_data(query=ticker_query)



import logging
from datetime import datetime

import pandas as pd
import pytest
import requests

import earningspy.generators.yahoo.time_series as time_series


class FakeResponse:

    def __init__(self, ok=True, status_code=200, payload=None, text=''):
        self.ok = ok
        self.status_code = status_code
        self._payload = payload
        self.text = text

    def json(self):
        if isinstance(self._payload, Exception):
            raise self._payload
        return self._payload


def _sample_ticker_frame():
    return pd.DataFrame(
        {
            'open': [100.0, 101.0],
            'high': [101.0, 102.0],
            'low': [99.0, 100.0],
            'close': [100.5, 101.5],
            'volume': [1000, 1100],
        },
        index=pd.to_datetime(['2024-01-02', '2024-01-03']),
    ).rename_axis('Date')


def test_get_one_ticker_logs_and_returns_none_on_request_exception(monkeypatch, caplog):
    def fake_get(*args, **kwargs):
        raise requests.RequestException("network down")

    monkeypatch.setattr(time_series.requests, 'get', fake_get)

    with caplog.at_level(logging.WARNING, logger=time_series.__name__):
        result = time_series.get_one_ticker('AAPL', start_date='2024-01-01', end_date='2024-01-31')

    assert result is None
    assert "Yahoo request failed for AAPL" in caplog.text


def test_get_one_ticker_logs_non_ok_response(monkeypatch, caplog):
    def fake_get(*args, **kwargs):
        return FakeResponse(ok=False, status_code=503, text='service unavailable')

    monkeypatch.setattr(time_series.requests, 'get', fake_get)

    with caplog.at_level(logging.WARNING, logger=time_series.__name__):
        result = time_series.get_one_ticker('AAPL', start_date='2024-01-01', end_date='2024-01-31')

    assert result is None
    assert "Yahoo request returned non-OK response for AAPL" in caplog.text
    assert "status=503" in caplog.text


def test_get_one_ticker_logs_malformed_payload(monkeypatch, caplog):
    payload = {'chart': {'result': None}}

    def fake_get(*args, **kwargs):
        return FakeResponse(ok=True, payload=payload)

    monkeypatch.setattr(time_series.requests, 'get', fake_get)

    with caplog.at_level(logging.WARNING, logger=time_series.__name__):
        result = time_series.get_one_ticker('AAPL', start_date='2024-01-01', end_date='2024-01-31')

    assert result is None
    assert "Yahoo payload parsing failed for AAPL" in caplog.text


def test_get_portfolio_skips_failed_ticker_and_keeps_valid_data(monkeypatch, caplog):
    sample = _sample_ticker_frame()

    def fake_get_one_ticker(asset, from_='3m', start_date=None, end_date=None):
        if asset == 'BAD':
            return None
        return sample.copy()

    monkeypatch.setattr(time_series, 'get_one_ticker', fake_get_one_ticker)
    monkeypatch.setattr(time_series, 'sleep', lambda _: None)
    monkeypatch.setattr(time_series, 'tqdm', lambda items: items)

    with caplog.at_level(logging.INFO, logger=time_series.__name__):
        portfolio = time_series.get_portfolio(['GOOD', 'BAD'], start_date='2024-01-01', end_date=datetime(2024, 1, 31).date())

    assert 'GOOD' in portfolio.columns
    assert 'BAD' not in portfolio.columns
    assert "Skipping BAD: no data returned from Yahoo fetch" in caplog.text


def test_get_portfolio_logs_malformed_data_skip(monkeypatch, caplog):
    sample = _sample_ticker_frame()
    original_prepare_data = time_series.prepare_data

    def fake_get_one_ticker(asset, from_='3m', start_date=None, end_date=None):
        return sample.copy()

    def fake_prepare_data(data, ticker):
        if ticker == 'BAD':
            return pd.DataFrame({'wrong': [1.0]}, index=pd.to_datetime(['2024-01-02'])).rename_axis('Date')
        return original_prepare_data(sample.copy(), ticker)

    monkeypatch.setattr(time_series, 'get_one_ticker', fake_get_one_ticker)
    monkeypatch.setattr(time_series, 'prepare_data', fake_prepare_data)
    monkeypatch.setattr(time_series, 'sleep', lambda _: None)
    monkeypatch.setattr(time_series, 'tqdm', lambda items: items)

    with caplog.at_level(logging.WARNING, logger=time_series.__name__):
        portfolio = time_series.get_portfolio(['BAD', 'GOOD'], start_date='2024-01-01', end_date='2024-01-31')

    assert 'GOOD' in portfolio.columns
    assert 'BAD' not in portfolio.columns
    assert "Skipping BAD: malformed data after preparation" in caplog.text


def test_get_portfolio_logs_and_raises_when_all_assets_fail(monkeypatch, caplog):
    monkeypatch.setattr(time_series, 'get_one_ticker', lambda *args, **kwargs: None)
    monkeypatch.setattr(time_series, 'tqdm', lambda items: items)

    with caplog.at_level(logging.ERROR, logger=time_series.__name__):
        with pytest.raises(ValueError, match="No valid assets found"):
            time_series.get_portfolio(['BAD1', 'BAD2'], start_date='2024-01-01', end_date='2024-01-31')

    assert "No valid assets found for Yahoo portfolio request" in caplog.text


def test_get_portfolio_avoids_repeated_merge_and_preserves_date_union(monkeypatch):
    frames = {
        'AAA': pd.DataFrame(
            {
                'open': [10.0, 11.0],
                'high': [10.5, 11.5],
                'low': [9.5, 10.5],
                'close': [10.1, 11.1],
                'volume': [100, 110],
            },
            index=pd.to_datetime(['2024-01-02', '2024-01-03']),
        ).rename_axis('Date'),
        'BBB': pd.DataFrame(
            {
                'open': [20.0, 21.0],
                'high': [20.5, 21.5],
                'low': [19.5, 20.5],
                'close': [20.2, 21.2],
                'volume': [200, 210],
            },
            index=pd.to_datetime(['2024-01-03', '2024-01-04']),
        ).rename_axis('Date'),
    }

    def fake_get_one_ticker(asset, from_='3m', start_date=None, end_date=None):
        return frames[asset].copy()

    monkeypatch.setattr(time_series, 'get_one_ticker', fake_get_one_ticker)
    monkeypatch.setattr(time_series, 'sleep', lambda _: None)
    monkeypatch.setattr(time_series, 'tqdm', lambda items: items)
    monkeypatch.setattr(
        time_series.pd,
        'merge',
        lambda *args, **kwargs: (_ for _ in ()).throw(AssertionError("pd.merge should not be called")),
    )

    portfolio = time_series.get_portfolio(['AAA', 'BBB', 'AAA'], start_date='2024-01-01', end_date='2024-01-31')

    assert list(portfolio.columns) == ['AAA', 'BBB']
    assert portfolio.index.tolist() == list(pd.to_datetime(['2024-01-02', '2024-01-03', '2024-01-04']))
    assert portfolio.loc[pd.Timestamp('2024-01-02'), 'AAA'] == pytest.approx(10.1)
    assert pd.isna(portfolio.loc[pd.Timestamp('2024-01-02'), 'BBB'])
    assert portfolio.loc[pd.Timestamp('2024-01-04'), 'BBB'] == pytest.approx(21.2)
